#    This file is part of Cruscoplanets GUI.
#
#   Cruscoplanets GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Cruscoplanets GUI is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Cruscoplanets GUI.  If not, see <http://www.gnu.org/licenses/>.
from PyQt5.QtWidgets import QTabWidget, QMainWindow, QApplication, QFileDialog, QMessageBox
from PyQt5.QtCore import Qt, QTranslator, QFile, QTextStream, QIODevice
from PyQt5.QtGui import QIcon
import os, sys
from .main_window_widgets import chart
from .settings import settings_form
from .settings import language_settings_form
from . import config
from .databases import database_models
from .databases import database_views
from . import resources


class TabWidget(QTabWidget):

	def __init__(self):
		super().__init__()
		self.birthChart = chart.BirthChart()
		self.transitChart = chart.TransitChart()
		self.addTab(self.birthChart, self.tr("Birth Chart"))
		self.addTab(self.transitChart, self.tr("Transit Chart"))
		
	def _updateGraphics(self):
		"""updates the graphics where settings are changed"""
		self.birthChart.chartForm.submitData.emit(self.birthChart.chartForm.chartdata())
		self.transitChart.chartForm.submitData.emit(self.transitChart.chartForm.chartdata())
		
	def newBirthChart(self):
		self.removeTab(0)
		self.birthChart = chart.BirthChart()
		self.insertTab(0, self.birthChart, self.tr("Birth Chart"))
		self.setCurrentIndex(0)
		
	def newTransitChart(self):
		self.removeTab(1)
		self.transitChart = chart.TransitChart()
		self.addTab(self.transitChart, self.tr("Transit Chart"))
		self.setCurrentIndex(1)
		
	def drawBirthChart(self, data):
		self.birthChart.chartForm.setDataFromDb(data)
		
	def drawTransitChart(self, data):
		self.transitChart.chartForm.setDataFromDb(data)

	def export(self, fileName):
		self.currentWidget().export(fileName)
	
	def currentData(self):
		return self.currentWidget().chartForm.chartdata(as_dict=True)


class MainWindow(QMainWindow):

	PERSONAL_DB_PATH = os.path.join(os.path.dirname(__file__), "databases", "personal.db")

	def __init__(self):
		super().__init__()
		self.setWindowIcon(QIcon(':/icon/images/cruscoplanets.png'))

		#calling configuration here makes sure that some initialization procedures are done, as creating the resource directories if they don't exist
		configuration = config.Configuration()

		self.setWindowTitle("Cruscoplanets GUI")
		self.tabWidget = TabWidget()
		self.setCentralWidget(self.tabWidget)
		self.actions = self.menuBar().addMenu(self.tr("Actions"))
		new_action = self.actions.addAction(self.tr("New"))
		open_action = self.actions.addAction(self.tr("Open_database"))
		save_action = self.actions.addAction(self.tr("Save_in_database"))
		self.databaseModel = database_models.PersonalDatabaseModel()
		self.peopleTableView = database_views.TableView(self.databaseModel.peopleTable)
		self.peopleTableView.setWindowTitle(self.tr("people"))
		self.peopleTableView.showData.connect(self._onShow)
		new_action.triggered.connect(self._onNew)
		open_action.triggered.connect(self.peopleTableView.show)
		save_action.triggered.connect(self._onAddToDatabase)


		export_action = self.actions.addAction(self.tr("Export"))
		export_action.triggered.connect(self._onExport)

		self.settings = self.menuBar().addMenu(self.tr("Settings"))
		graphic_settings = self.settings.addAction(self.tr("Graphics"))
		language_settings = self.settings.addAction(self.tr("Language"))
		self.about = self.menuBar().addMenu(self.tr("About"))
		about = self.about.addAction(self.tr("About cruscoplanets GUI"))
		about.triggered.connect(self._onAbout)

		self.settingsWindow = settings_form.SettingsWindow()
		graphic_settings.triggered.connect(self.settingsWindow.show)
		self.settingsWindow.optionsChanged.connect(self.tabWidget._updateGraphics)
		self.languageSettingsWindow = language_settings_form.LanguageSettingsWindow()
		self.languageSettingsWindow.chartLanguageUpdated.connect(self.tabWidget._updateGraphics)
		language_settings.triggered.connect(self.languageSettingsWindow.show)
		
	def _onOpen(self):
		self.peopleTableView.show()

	def _onShow(self, data):
		if self.tabWidget.currentIndex() == 0:
			self.tabWidget.drawBirthChart(data)
		elif self.tabWidget.currentIndex() == 1:
			self.tabWidget.drawTransitChart(data)

	def _onNew(self):
		if self.tabWidget.currentIndex() == 0:
			self.tabWidget.newBirthChart()
		elif self.tabWidget.currentIndex() == 1:
			self.tabWidget.newTransitChart()

	def _onAddToDatabase(self):
		data = self.tabWidget.currentData()
		print(data)
		for key in ('transitdtstring', 'transitlocationid'):#if the transit tab is open we save only the information related to birth
			if key in data.keys():
				del data[key]
		success = self.peopleTableView.tableData.addRow(data)
		if success:
			message = QMessageBox.information(self, self.tr("adding_success"), self.tr("the_new_birth_chart_has_been_successfully_added_to_the_database"))
		else:
			message = QMessageBox.critical(self, self.tr("adding_failure"), self.tr("the_new_birth_chart_has_not_been_successfully_added_to_the_database"))
			

	def _onAbout(self):
		message = QMessageBox()
		message.setWindowTitle(self.tr("about_cruscoplanets_GUI"))
		message.setTextFormat(Qt.MarkdownText)
		message.setIcon(QMessageBox.Information)
		about_file_path = config.Configuration().aboutFilePath()
		thefile = QFile(about_file_path)
		thefile.open(QIODevice.ReadOnly)
		thetext = QTextStream(thefile).readAll()
		thefile.close()
		message.setText(thetext)
		message.exec_()
		

	def _onExport(self):
		dialog = QFileDialog(self)
		dialog.setFileMode(QFileDialog.AnyFile)
		dialog.setDirectory(os.path.expanduser('~'))
		dialog.setNameFilter(self.tr("Image Files (*.png *.jpg *.svg)"))
		fileName, ok = dialog.getSaveFileName(self, self.tr("export_file"))
		if ok:
			self.tabWidget.export(fileName)

	def showEvent(self, event):
		"""Overridden to make the window occupy the whole screen"""
		super().showEvent(event)
		geometry = self.windowHandle().screen().availableGeometry()
		self.setGeometry(geometry)

		
	def closeEvent(self, event):
		self.settingsWindow.close()
		self.languageSettingsWindow.close()
		self.peopleTableView.close()
		self.tabWidget.birthChart.chartForm._closeLocationSelectors()
		self.tabWidget.transitChart.chartForm._closeLocationSelectors()
		super().close()

def main():
	app = QApplication(sys.argv)

	gui_language = config.Configuration().GUI_DEFAULT_LANGUAGE
	translator = QTranslator()
#	translator.load(os.path.join("translations", "tr_%s"%gui_language))
	translator.load(os.path.join(":/translations/translations/tr_%s.qm"%gui_language))
	app.installTranslator(translator)

	window = MainWindow()
	window.show()
	sys.exit(app.exec())

if __name__ == '__main__':
	main()



#    This file is part of Cruscoplanets GUI.
#
#   Cruscoplanets GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Cruscoplanets GUI is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Cruscoplanets GUI.  If not, see <http://www.gnu.org/licenses/>.
from PyQt5.QtWidgets import QWidget, QDateEdit, QTimeEdit, QHBoxLayout, QVBoxLayout, QGridLayout, QRadioButton, QCheckBox, QSpinBox
from PyQt5.QtCore import QDate, QTime, pyqtSignal, Qt


class TimeOptions(QWidget):

	optionsChanged = pyqtSignal()
	
	def __init__(self):
		super().__init__()
		
		buttonslayout = QGridLayout()
		self.localTimeButton = QRadioButton(self.tr('local time'), self)
		self.localTimeButton.setChecked(True)
		self.UTCTimeButton = QRadioButton(self.tr('UTC time'), self)
		buttonslayout.addWidget(self.localTimeButton, 0, 0)
		buttonslayout.addWidget(self.UTCTimeButton, 1, 0)
		self.unknownTimeButton = QCheckBox(self.tr('unknown time'), self)
		buttonslayout.addWidget(self.unknownTimeButton, 0, 1)
		
		self.localTimeButton.toggled.connect(self.optionsChanged.emit)
		self.UTCTimeButton.toggled.connect(self.optionsChanged.emit)
		self.unknownTimeButton.stateChanged.connect(self.optionsChanged.emit)
		self.setLayout(buttonslayout)
	
	@property
	def timeType(self):
		return 'l' if self.localTimeButton.isChecked() else 'u'
	
	@property
	def unknownTimeChar(self):
		return '!' if self.unknownTimeButton.isChecked() else '-'
		

class CarryingSpinBox(QSpinBox):

	ACCEPTED_KEYS = (Qt.Key_0, Qt.Key_1, Qt.Key_2, Qt.Key_3, Qt.Key_4, Qt.Key_5, Qt.Key_6, Qt.Key_7, Qt.Key_8, Qt.Key_9, Qt.Key_Backspace, Qt.Key_Delete)

	carry = pyqtSignal(int)
	
	def __init__(self, maxValue, minValue, suffix):
		super().__init__(maximum = maxValue, minimum = minValue)
		self.setSuffix(suffix)
		self.setWrapping(True)
		self._oldValue = self.value()
		self.valueChanged.connect(self._adjustValue)

	@property
	def maximumDigitsAllowed(self):
		return len(str(self.maximum()-1))

	def valueFromText(self, text):
		digits = filter(lambda char: char.isdigit(), text)
		value = ''.join(digits)
		return int(value)
	
	def textFromValue(self, value):
		return "%02d"%value

	def _adjustValue(self):
		if self._oldValue == self.maximum() and self.value() == self.minimum():
			self.carry.emit(1)
		elif self._oldValue == self.minimum() and self.value() == self.maximum():
			self.carry.emit(-1)
		self._oldValue = self.value()
		
			
	def keyPressEvent(self, event):
		self._oldValue = int((self.maximum()-self.minimum())/2)#so we avoid carrying
		super().keyPressEvent(event) #datetime doesn't change
		
		if len(self.cleanText()) > self.maximumDigitsAllowed:#this could happen if the user continuously inserts 0 chars
			self.setValue(0)
			
	def _acceptCarry(self, carry: int):
		self.setValue(self.value() + carry)
		
		
class TimeSpinBox(CarryingSpinBox):

	def __init__(self, maxvalue, suffix):
		super().__init__(maxvalue, 0, suffix)


class TimeFormatter(QWidget):

	dateCarry = pyqtSignal(int)
	timeChanged = pyqtSignal()

	def __init__(self, mytime: QTime = QTime.currentTime()):
		super().__init__()
		layout = QHBoxLayout()
		layout.setSpacing(0)
		self.hoursWidget = TimeSpinBox(23, 'h')
		self.hoursWidget.setValue(mytime.hour())
		self.minutesWidget = TimeSpinBox(59, 'm')
		self.minutesWidget.setValue(mytime.minute())
		self.secondsWidget = TimeSpinBox(59, 's')
		self.secondsWidget.setValue(mytime.second())

		self.secondsWidget.carry.connect(lambda unit: self.minutesWidget._acceptCarry(unit))
		self.minutesWidget.carry.connect(lambda unit: self.hoursWidget._acceptCarry(unit))
		self.hoursWidget.carry.connect(lambda unit: self.dateCarry.emit(unit))
		
		self.secondsWidget.valueChanged.connect(lambda: self.timeChanged.emit())
		self.minutesWidget.valueChanged.connect(lambda: self.timeChanged.emit())
		self.hoursWidget.valueChanged.connect(lambda: self.timeChanged.emit())
		
		layout.addWidget(self.hoursWidget)
		layout.addWidget(self.minutesWidget)
		layout.addWidget(self.secondsWidget)
		self.setLayout(layout)

	@property
	def values(self):
		return (self.hoursWidget.value(), self.minutesWidget.value(), self.secondsWidget.value())
		
	def setTime(self, hours, minutes, seconds):
		self.hoursWidget.setValue(hours)
		self.minutesWidget.setValue(minutes)
		self.secondsWidget.setValue(seconds)
		
		
class DaySpinBox(CarryingSpinBox):

	DAYS_FOR_MONTHS = (31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31)

	def __init__(self):
		super().__init__(31, 1, 'd')
		
	def _adaptToYearAndMonth(self, isLeap: bool, month: int):
		month %= 12
		new_maximum = self.__class__.DAYS_FOR_MONTHS[month-1]
		if month == 2 and isLeap:
			new_maximum = self.__class__.DAYS_FOR_MONTHS[month-1]+1
		self._oldValue = int((new_maximum-self.minimum())/2)#in this way we avoid carrying
		if self.value() > new_maximum:
			self.setValue(new_maximum)
		self.setMaximum(new_maximum)

	def setToMaximum(self, month_value):
		"""sets the day to the maximum value allowed for the month. Useful where month has negative carry"""
		self.setValue(self.__class__.DAYS_FOR_MONTHS[month_value-1])
		
		
class MonthSpinBox(CarryingSpinBox):

	adaptDay = pyqtSignal(int)

	def __init__(self):
		super().__init__(12, 1, 'm')
		
	def _acceptCarry(self, unit):
		super()._acceptCarry(unit)
		if unit == -1:
			self.adaptDay.emit(self.value())
		
		
class YearSpinBox(CarryingSpinBox):

	def __init__(self):
		super().__init__(10000, 0, 'y')
	
	def isLeap(self):
		value = self.value()
		return (value % 4 == 0) and ((value % 100 != 0) or (value % 400) == 0)
		

class DateFormatter(QWidget):

	dateChanged = pyqtSignal()

	def __init__(self, mydate: QDate = QDate.currentDate()):
		super().__init__()
		layout = QHBoxLayout()
		layout.setSpacing(0)
		self.yearsWidget = YearSpinBox()
		self.yearsWidget.setValue(mydate.year())
		self.monthsWidget = MonthSpinBox()
		self.monthsWidget.setValue(mydate.month())
		self.daysWidget = DaySpinBox()
		self.daysWidget.setValue(mydate.day())
		self.daysWidget._adaptToYearAndMonth(self.yearsWidget.isLeap(), self.monthsWidget.value())

		self.daysWidget.carry.connect(lambda unit: self.monthsWidget._acceptCarry(unit))
		self.monthsWidget.carry.connect(lambda unit: self.yearsWidget._acceptCarry(unit))
		
		self.yearsWidget.valueChanged.connect(lambda year: self.daysWidget._adaptToYearAndMonth(self.yearsWidget.isLeap(), self.monthsWidget.value()))
		self.monthsWidget.valueChanged.connect(lambda month: self.daysWidget._adaptToYearAndMonth(self.yearsWidget.isLeap(), month))
		self.monthsWidget.adaptDay.connect(lambda value: self.daysWidget.setToMaximum(value))
		
		self.yearsWidget.valueChanged.connect(lambda: self.dateChanged.emit())
		self.monthsWidget.valueChanged.connect(lambda: self.dateChanged.emit())
		self.daysWidget.valueChanged.connect(lambda: self.dateChanged.emit())
		
		layout.addWidget(self.yearsWidget)
		layout.addWidget(self.monthsWidget)
		layout.addWidget(self.daysWidget)
		self.setLayout(layout)
	
	def _carryOnDate(self, carry):
		self.daysWidget.setValue(self.daysWidget.value() + carry)
		
	def values(self):
		return (self.yearsWidget.value(), self.monthsWidget.value(), self.daysWidget.value())
	
	def setDate(self, year, month, day):
		self.yearsWidget.setValue(year)
		self.monthsWidget.setValue(month)
		self.daysWidget.setValue(day)


class DateTimeFormatter(QWidget):

	datetimeChanged = pyqtSignal(str)

	def __init__(self, mydate: QDate = QDate.currentDate(), mytime: QTime = QTime.currentTime()):
		super().__init__()
		firstlayout = QVBoxLayout()

		datetimelayout = QHBoxLayout()
		self.dateWidget = DateFormatter(mydate)
		self.dateWidget.dateChanged.connect(self._onDateTimeChanged)
		datetimelayout.addWidget(self.dateWidget)
		self.timeWidget = TimeFormatter(mytime)
		datetimelayout.addWidget(self.timeWidget)
		firstlayout.addLayout(datetimelayout)
		self.timeWidget.dateCarry.connect(lambda unit: self.dateWidget._carryOnDate(unit))
		self.timeWidget.timeChanged.connect(self._onDateTimeChanged)
		
		self.timeOptionsWidget = TimeOptions()
		self.timeOptionsWidget.optionsChanged.connect(self._onDateTimeChanged)
		firstlayout.addWidget(self.timeOptionsWidget)
		self.setLayout(firstlayout)

	def _onDateTimeChanged(self):
		self.datetimeChanged.emit(self.dateTimeString())

	def dateTimeString(self):
		timetype = self.timeOptionsWidget.timeType
		ret_str = timetype + '%04d-%02d-%02d'%(self.dateWidget.values())
		unknown_hour_char = self.timeOptionsWidget.unknownTimeChar
		ret_str += unknown_hour_char
		ret_str += '%02d-%02d-%02d'%(self.timeWidget.values)
		return ret_str
	
	def setDateTime(self, event):
		if event.cruscoplanetsString[0] == 'u':
			self.dateWidget.setDate(event.utcdatetime.year, event.utcdatetime.month, event.utcdatetime.day)
			self.timeWidget.setTime(event.utcdatetime.hour, event.utcdatetime.minute, event.utcdatetime.second)
			self.timeOptionsWidget.UTCTimeButton.setChecked(True)
		else:
			self.dateWidget.setDate(event.localdatetime.year, event.localdatetime.month, event.localdatetime.day)
			self.timeWidget.setTime(event.localdatetime.hour, event.localdatetime.minute, event.localdatetime.second)
			self.timeOptionsWidget.localTimeButton.setChecked(True)
		if event.isTimeUnknown:
			self.timeOptionsWidget.unknownTimeButton.setChecked(True)
		else:
			self.timeOptionsWidget.unknownTimeButton.setChecked(False)



#    This file is part of Cruscoplanets GUI.
#
#   Cruscoplanets GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Cruscoplanets GUI is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Cruscoplanets GUI.  If not, see <http://www.gnu.org/licenses/>.
from PyQt5.QtWidgets import QWidget, QVBoxLayout, QPushButton, QSizePolicy
from PyQt5.QtCore import pyqtSignal
from .event_form import EventForm
from .location_searcher import LocationSearcher
from .personal_data_form import PersonalDataForm
from ..config import Configuration


class ChartForm(QWidget):

	submitData = pyqtSignal(tuple)

	def __init__(self, name: str = None, surname: str = None, sex: int = None, birthtime: str = None, birthlocation_id: int = None):
		super().__init__()
		self.configuration = Configuration()
		
		self.baseLayout = QVBoxLayout()
		
		self.personalDataForm = PersonalDataForm(name, surname, sex)
		self.baseLayout.addWidget(self.personalDataForm)
		
		#place and time for birth event is required in any case, so we put them here in the base class:
		self.birthEventForm = EventForm(birthtime, birthlocation_id)
		self.birthEventForm.eventChanged.connect(lambda: self.submitData.emit(self.chartdata()))
		self.baseLayout.addWidget(self.birthEventForm)
		
		self.setLayout(self.baseLayout)
		self.setSizePolicy(QSizePolicy.Maximum, QSizePolicy.Maximum)
		
	def _closeLocationSelectors(self):
		self.birthEventForm.locationSearcher._closeLocationSelector()
		
	def setDataFromDb(self, data_dict):
		self.personalDataForm.nameField.setText(data_dict['name'])
		self.personalDataForm.surnameField.setText(data_dict['surname'])
		self.personalDataForm.setSex(data_dict['sex'])
		self.birthEventForm.setEvent(data_dict['birthdtstring'], data_dict['birthlocationid'])
		
		
class BirthChartForm(ChartForm):

	def __init__(self, name: str = None, surname: str = None, sex: int = None, birthtime: str = None, birthlocation_id: int = None):
		super().__init__(name, surname, sex, birthtime, birthlocation_id)
		
	def voidchartdata(self):#returns a chart of a not specified person, useful when opening the program
		return (
			self.birthEventForm.datetime, self.birthEventForm.location_id, 
			"here", "and now", self.personalDataForm.sex, 
			self.configuration.CHART_DEFAULT_LANGUAGE

		)
		
	def chartdata(self, as_dict = False):
		nameValue = self.personalDataForm.nameField.text() if self.personalDataForm.nameField.text() != '' else "--"
		surnameValue = self.personalDataForm.surnameField.text() if self.personalDataForm.surnameField.text() != '' else "--"
		if not as_dict:
			return (
				self.birthEventForm.datetime, self.birthEventForm.location_id, 
				nameValue, surnameValue, self.personalDataForm.sex, 
				self.configuration.CHART_DEFAULT_LANGUAGE
			)
		else:
			return {
				'birthdtstring': self.birthEventForm.datetime, 
				'birthlocationid': self.birthEventForm.location_id, 
				'name': nameValue, 
				'surname': surnameValue, 
				'sex': self.personalDataForm.sex, 
			}
		
		
class TransitChartForm(ChartForm):

	def __init__(
		self, name: str = None, surname: str = None, sex: int = None, 
		birthtime: str = None, birthlocation_id: int = None, 
		transittime: str = None, transitlocation_id: int = None, 
	):
		super().__init__(name, surname, sex, birthtime, birthlocation_id)
		self.transitEventForm = EventForm(transittime, transitlocation_id)
		self.transitEventForm.eventChanged.connect(lambda: self.submitData.emit(self.chartdata()))
		self.baseLayout.addWidget(self.transitEventForm)
		
	def voidchartdata(self):#returns a chart of a not specified person, useful when opening the program
		return (
			self.birthEventForm.datetime, self.birthEventForm.location_id, 
			self.transitEventForm.datetime, self.transitEventForm.location_id,
			"here", "and now", self.personalDataForm.sex, 
			self.configuration.CHART_DEFAULT_LANGUAGE
		)
		
	def chartdata(self, as_dict = False):
		nameValue = self.personalDataForm.nameField.text() if self.personalDataForm.nameField.text() != '' else "--"
		surnameValue = self.personalDataForm.surnameField.text() if self.personalDataForm.surnameField.text() != '' else "--"
		if not as_dict:
			return (
				self.birthEventForm.datetime, self.birthEventForm.location_id, 
				self.transitEventForm.datetime, self.transitEventForm.location_id,
				nameValue, surnameValue, self.personalDataForm.sex, 
				self.configuration.CHART_DEFAULT_LANGUAGE
			)
		else:
			return {
				'birthdtstring': self.birthEventForm.datetime, 
				'birthlocationid': self.birthEventForm.location_id, 
				'transitdtstring': self.transitEventForm.datetime, 
				'transitlocationid': self.transitEventForm.location_id, 
				'name': nameValue, 
				'surname': surnameValue, 
				'sex': self.personalDataForm.sex, 
			}
		
	def _closeLocationSelectors(self):
		self.birthEventForm.locationSearcher._closeLocationSelector()
		self.transitEventForm.locationSearcher._closeLocationSelector()



#    This file is part of Cruscoplanets GUI.
#
#   Cruscoplanets GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Cruscoplanets GUI is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Cruscoplanets GUI.  If not, see <http://www.gnu.org/licenses/>.
import PyQt5.QtWidgets as qtw
import PyQt5.QtCore as qtc
import PyQt5.QtGui as qtg


class VoidWidget(qtw.QWidget):
	
	def __init__(self):
		super().__init__()
		palette = qtg.QPalette()
		palette.setColor(qtg.QPalette.Window, qtg.QColor("white"))
		self.setAutoFillBackground(True)
		self.setPalette(palette)


class LayoutItem(qtw.QLayoutItem):
	
	def __init__(self, alignment):
		super().__init__(alignment)
		
	def expandingDirections(self):
		return 0


class GridLayout(qtw.QGridLayout):

	def __init__(self):
		super().__init__()
		self.setHorizontalSpacing(2)
	
	def	addItem(self, item, row, column, rowSpan = 1, columnSpan = 1, alignment = qtc.Qt.AlignLeft):
		item = LayoutItem(item.alignment())
		super().addItem(item, row, column, rowSpan, columnSpan, alignment)
	
	def fillSpaces(self, rows = None, columns = None):
		rows = rows if rows != None else self.rowCount()
		columns = columns if columns != None else self.columnCount()
		for i in range(rows):
			for j in range(columns):
				if self.itemAtPosition(i, j) == None:
					void_widget = VoidWidget()
					self.addWidget(void_widget, i, j)
					void_widget.show()
					
	def width(self):
		ret = 0
		if self.rowCount() > 0:
			for i in range(self.columnCount()):
				item = self.itemAtPosition(0, i)
				ret += item.widget().width()
			hspacings = self.columnCount() -1
			ret += hspacings*self.horizontalSpacing()
			left, top, right, bottom = self.getContentsMargins()
		return ret + left + right
					
	def height(self):
		ret = 0
		if self.columnCount() > 0:
			for i in range(self.rowCount()):
				item = self.itemAtPosition(1, 0)
				ret += item.widget().height()
			vspacings = self.rowCount() -1
			ret += vspacings*self.verticalSpacing()
			left, top, right, bottom = self.getContentsMargins()
		return ret + top + bottom
	
	def getMaxWidth(self, column_index):
		widths = []
		for i in range(self.rowCount()):
			item = self.itemAtPosition(i, column_index)
			if not item.widget().isVisible():
				widths.append(0)
			else:
				georect = item.geometry()
				widths.append(georect.width())
		return max(widths)

	def setWidth(self, column_index, width):
		for i in range(self.rowCount()):
			item = self.itemAtPosition(i, column_index)
			georect = item.geometry()
			georect.setWidth(width)
			item.setGeometry(georect)
			item.widget().setFixedWidth(georect.width())
			item.widget().setSizePolicy(qtw.QSizePolicy.Fixed, qtw.QSizePolicy.Fixed)
	
	def getMaxHeight(self, row_index):
		heights = []
		for i in range(self.columnCount()):
			item = self.itemAtPosition(row_index, i)
			if not item.widget().isVisible():
				heights.append(0)
			else:
				georect = item.geometry()
				heights.append(georect.height())
		return max(heights)

	def setHeight(self, row_index, height):
		for i in range(self.columnCount()):
			item = self.itemAtPosition(row_index, i)
			georect = item.geometry()
			georect.setHeight(height)
			item.setGeometry(georect)
			item.widget().setFixedHeight(georect.height())
			item.widget().setSizePolicy(qtw.QSizePolicy.Fixed, qtw.QSizePolicy.Fixed)




#    This file is part of Cruscoplanets GUI.
#
#   Cruscoplanets GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Cruscoplanets GUI is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Cruscoplanets GUI.  If not, see <http://www.gnu.org/licenses/>.
from PyQt5.QtWidgets import QWidget, QVBoxLayout
from PyQt5.QtCore import pyqtSignal
from .location_searcher import LocationSearcher
from .datetime_formatter import DateTimeFormatter
from ..config import Configuration
import cruscoplanets.cmdline


class EventForm(QWidget):
	"""Widget for entering the data of a cruscoplanets event.
	
	Args:
		datetimestring: the string in cruscoplanet format representing the initial date and time that the EventForm widget will display. If None, the default date and time string will be used.
		location_id: the integer representing the initial location that the EventForm widget will display. If None, the default location id will be used.
	"""

	eventChanged = pyqtSignal(str, int)

	def __init__(self, datetimestring: str = None, location_id: int = None):
		super().__init__()
		self.configuration = Configuration()
		self.location_id = location_id if location_id != None else self.configuration.DEFAULT_LOCATION
		if datetimestring != None:
			self.datetime = datetimestring
		else:
			if self.configuration.DEFAULT_DATETIME_STRING != 'now':
				self.datetime = self.configuration.DEFAULT_DATETIME_STRING
			else:
				self.datetime = cruscoplanets.cmdline.buildCurrentEvent(self.location_id).cruscoplanetsString

		self.locationSearcher = LocationSearcher()
		self.locationSearcher.locationChosen.connect(lambda loc_id: self._updateLocation(loc_id))
		self.datetimeFormatter = DateTimeFormatter()
		self.datetimeFormatter.datetimeChanged.connect(lambda dtstring: self._updateDatetime(dtstring))
		self.setEvent(self.datetime, self.location_id)
		
		layout = QVBoxLayout()
		layout.addWidget(self.locationSearcher)
		layout.addWidget(self.datetimeFormatter)
		self.setLayout(layout)

	def setEvent(self, datetime, location_id):
		self.event = cruscoplanets.cmdline.buildEvent(datetime, location_id)
		
		self.locationSearcher.setLocation(location_id)
		self.location_id = location_id
		self.datetimeFormatter.setDateTime(self.event)
		self.eventChanged.emit(datetime, location_id)
	
	def _updateLocation(self, loc_id):
		self.location_id = loc_id
		self.eventChanged.emit(self.datetime, self.location_id)

	def _updateDatetime(self, dtstring):
		self.datetime = dtstring
		self.eventChanged.emit(self.datetime, self.location_id)



#    This file is part of Cruscoplanets GUI.
#
#   Cruscoplanets GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Cruscoplanets GUI is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Cruscoplanets GUI.  If not, see <http://www.gnu.org/licenses/>.
from PyQt5 import QtSvg
from PyQt5.QtWidgets import QSizePolicy, QMessageBox
from PyQt5.QtCore import Qt, QSize
from cruscoplanets.cmdline import getImagesSize


class SvgWidget(QtSvg.QSvgWidget):

	def __init__(self):
		super().__init__()
		width, height = getImagesSize()
		self.ratio = width/height

	def resizeEvent(self, event):
		super().resizeEvent(event)
		new_width = self.width()
		new_height = self.height() 
		if new_width/new_height > self.ratio:
			new_width = new_height * self.ratio
		elif new_width/new_height < self.ratio:
			new_height = new_width/self.ratio
		self.resize(new_width, new_height)



#    This file is part of Cruscoplanets GUI.
#
#   Cruscoplanets GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Cruscoplanets GUI is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Cruscoplanets GUI.  If not, see <http://www.gnu.org/licenses/>.
from PyQt5.QtWidgets import QLineEdit, QWidget, QMainWindow, QHBoxLayout, QDateTimeEdit, QComboBox, QCheckBox, QGridLayout, QTableView, QMessageBox, QStyledItemDelegate, QLabel, QStyle, QProxyStyle
from PyQt5.QtCore import Qt, QObject, pyqtSignal, QDate, QTime, QDateTime
from PyQt5.QtGui import QColor, QPalette, QPen
from PyQt5.QtSql import QSqlDatabase, QSqlTableModel
from ..main_window_widgets import location_searcher
import cruscoplanets.cmdline
import os.path
from .. import config


class Delegate:

	updated = pyqtSignal()

	def __init__(self):
		self._row = None
		self._column = None
		
	def setRow(self, index: int):
		self._row = index
		
	def setColumn(self, index: int):
		self._column = index
		

class LocationDatum(location_searcher.LocationSearcher, Delegate):
	
	def __init__(self):
		super().__init__()
		self.layout().setContentsMargins(0, 0, 0, 0)
		self.locationChosen.connect(self.updated.emit)
		
	def value(self):
		return self.selectedId


class IdLabel(QWidget, Delegate):

	def __init__(self, text):
		super().__init__()
		self.label = QLabel(text)
		layout = QHBoxLayout()
		layout.setContentsMargins(0, 0, 0, 0)
		layout.addWidget(self.label)
		self.setLayout(layout)
		
	def value(self):
		return int(self.label.text())
		


class LineEdit(QWidget, Delegate):

	def __init__(self):
		super().__init__()
		self.lineEdit = QLineEdit()
		layout = QHBoxLayout()
		layout.setContentsMargins(0, 0, 0, 0)
		layout.addWidget(self.lineEdit)
		self.setLayout(layout)
		self.lineEdit.textChanged.connect(self.updated.emit)
	
	def text(self):
		return self.lineEdit.text()
	
	def value(self):
		return self.lineEdit.text()
	
	def setText(self, value):
		self.lineEdit.setText(value)


class SexComboBox(QWidget, Delegate):

	def __init__(self):
		super().__init__()
		self._combobox = QComboBox()
		self.addItem(self.tr("male"))
		self.addItem(self.tr("female"))
		self.addItem(self.tr("undefined"))
		layout = QHBoxLayout()
		layout.setContentsMargins(0, 0, 0, 0)
		layout.addWidget(self._combobox)
		self.setLayout(layout)
		self._combobox.currentIndexChanged.connect(self.updated.emit)
		
	def addItem(self, value):
		self._combobox.addItem(value)
		
	def value(self):
		index = self._combobox.currentIndex()
		if index == 2:
			index = - 1
		return index
		
	def setCurrentIndex(self, sex_int):
		index = sex_int
		if index == -1:
			index = 2
		self._combobox.setCurrentIndex(index)


class TimeTypeComboBox(QComboBox):

	def __init__(self):
		super().__init__()
		self.addItem(self.tr("local_time"))
		self.addItem(self.tr("utc_time"))

	def setValue(self, value):
		if type(value) == str:
			value = True if value == 'l' else False
		if value == True:
			self.setCurrentIndex(0)
		elif value == False:
			self.setCurrentIndex(1)
			
	def currentValue(self):
		if self.currentIndex() == 0:
			return 'l'
		else:
			return 'u'


class DatetimeSetterError(Exception):

	def __init__(self):
		super().__init__()
	
	def __str__(self):
		return "The value 'now' can not be stored in the database as datetimestring."
		
		
class DatetimeSetter(QWidget, Delegate):

	def __init__(self):
		super().__init__()

		self.timeType = TimeTypeComboBox()
		self.datetime = QDateTimeEdit()
		self.datetime.setCalendarPopup(True)
		self.datetime.setDisplayFormat("yyyy/MM/dd hh:mm:ss")
		self.timeUnknown = QCheckBox(self.tr("unknown_time"))
		
		layout = QHBoxLayout()
		layout.setSpacing(0)
		layout.setContentsMargins(0, 0, 0, 0)
		layout.addWidget(self.timeType)
		layout.addWidget(self.datetime)
		layout.addWidget(self.timeUnknown)
		self.setLayout(layout)
		self.timeType.currentIndexChanged.connect(self.updated.emit)
		self.datetime.dateTimeChanged.connect(self.updated.emit)
		self.timeUnknown.clicked.connect(self.updated.emit)

	def setDatetimestring(self, dtstring):
		if dtstring == 'now':
			raise DatetimeSetterError
		
		is_time_local, is_time_known, year, month, day, hour, minute, second = cruscoplanets.cmdline.parseDatetimeString(dtstring)
		if hour == None:
			hour = 12
		if minute == None:
			minute = 0
		if second == None:
			second = 0
		date = QDate(year, month, day)
		time = QTime(hour, minute, second)
		self.timeType.setValue(is_time_local)
		self.datetime.setDateTime(QDateTime(date, time))
		self.timeUnknown.setChecked(not is_time_known)

	def value(self):
		return self.datetimeString()
	
	def datetimeString(self):
		ret_str = self.timeType.currentValue()
		ret_str += "%04d-%02d-%02d"%(self.datetime.date().year(), self.datetime.date().month(), self.datetime.date().day())
		unknown_time_char = '!' if self.timeUnknown.isChecked() else '-'
		ret_str += unknown_time_char
		ret_str += "%02d-%02d-%02d"%(self.datetime.time().hour(), self.datetime.time().minute(), self.datetime.time().second())
		return ret_str




#    This file is part of Cruscoplanets GUI.
#
#   Cruscoplanets GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Cruscoplanets GUI is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Cruscoplanets GUI.  If not, see <http://www.gnu.org/licenses/>.
import PyQt5.QtWidgets as qtw
import PyQt5.QtCore as qtc
import PyQt5.QtGui as qtg
from . import database_models
from . import delegates
from . import database_filters
from . import resizable_grid
from .. import resources

class Header(qtw.QWidget):

	sort = qtc.pyqtSignal(str, bool)

	def __init__(self, column):
		super().__init__()
		self.column = column
		layout = qtw.QHBoxLayout()
		layout.setSpacing(0)
		layout.setContentsMargins(0, 0, 0, 0)
		self.label = qtw.QLabel(self.column.header())
		layout.addWidget(self.label, qtc.Qt.AlignLeft)
		self.sortRight = qtw.QPushButton()
		self.sortRight.setIcon(self.style().standardIcon(qtw.QStyle.SP_ArrowDown))
		self.sortRight.setFixedWidth(self.sortRight.height())
		layout.addWidget(self.sortRight, qtc.Qt.AlignRight)
		self.sortReverse = qtw.QPushButton()
		self.sortReverse.setIcon(self.style().standardIcon(qtw.QStyle.SP_ArrowUp))
		self.sortReverse.setGeometry(qtc.QRect(0, 0, 20, 20))
		layout.addWidget(self.sortReverse, qtc.Qt.AlignRight)
		self.sortReverse.hide()
		self.setLayout(layout)
		
		self.sortRight.clicked.connect(self._onsortRightClicked)
		self.sortReverse.clicked.connect(self._onsortReverseClicked)

		self.sortRight.setSizePolicy(qtw.QSizePolicy.Fixed, qtw.QSizePolicy.Fixed)
		self.sortReverse.setSizePolicy(qtw.QSizePolicy.Fixed, qtw.QSizePolicy.Fixed)
		palette = qtg.QPalette()
		palette.setColor(qtg.QPalette.Window, qtg.QColor("white"))
		self.setAutoFillBackground(True)
		self.setPalette(palette)
		
	def _onsortRightClicked(self):
		self.sortRight.hide()
		self.sortReverse.show()
		self.sort.emit(self.column.name(), False)
		
	def _onsortReverseClicked(self):
		self.sortReverse.hide()
		self.sortRight.show()
		self.sort.emit(self.column.name(), True)
	
	def showEvent(self, event):
		super().showEvent(event)
		self.sortRight.setFixedWidth(self.sortRight.height())
		self.sortReverse.setFixedWidth(self.sortRight.height())
		

class Row(qtc.QObject):

	updated = qtc.pyqtSignal(dict)
	showData = qtc.pyqtSignal(dict)

	def __init__(self, record):
		super().__init__()
		self.record = record
		self.selectCheckBox = qtw.QCheckBox()
		self.showButton = qtw.QPushButton(qtg.QIcon(':/icon/images/show.png'), '')
		self.dataWidgets = tuple(self.buildWidget(field) for field in record.fields)
		self.widgets = (self.selectCheckBox,) + self.dataWidgets + (self.showButton,)
		self.showButton.clicked.connect(self._onShow)
	
	def buildWidget(self, field):
		if field.__class__ == database_models.SexField:
			ret_widget = delegates.SexComboBox()
			ret_widget.setCurrentIndex(field.value())
		elif field.__class__ == database_models.DtStringField:
			ret_widget = delegates.DatetimeSetter()
			ret_widget.setDatetimestring(field.value())
		elif field.__class__ == database_models.LocationField:
			ret_widget = delegates.LocationDatum()
			ret_widget.setLocation(field.value())
		elif field.__class__ == database_models.IdField:
			ret_widget = delegates.IdLabel(str(field.value()))
		else:
			ret_widget = delegates.LineEdit()
			ret_widget.setText(field.value())
			
		palette = qtg.QPalette()
		ret_widget.updated.connect(self.emitUpdate)
		return ret_widget
		
	def _onShow(self):
		data = {}
		for i in range(len(self.record.fields)):
			data.update({self.record.fields[i].name(): self.dataWidgets[i].value()})
		self.showData.emit(data)
		
	def emitUpdate(self):
		data = {}
		for i in range(len(self.record.fields)):
			data.update({self.record.fields[i].name(): self.dataWidgets[i].value()})
		self.updated.emit(data)
		
	def isSelected(self):
		return self.selectCheckBox.isChecked()
		
	def setSelected(self, value: bool):
		return self.selectCheckBox.setChecked(value)
		

class RowsSequence:

	def __init__(self, rows: tuple, columns: tuple):
		self.rows = rows
		self.columns = columns
		
	def __iter__(self):
		return iter(self.rows)
		
	def __getitem__(self, index):
		return self.rows[index]
		
	def iterateOnIndex(self, index):
		index = self.columns[index].index()
		for row in self.rows:
			yield row.widgets[index]
	
	def maxSizesByIndex(self, index):
		widths = [widget.width() for widget in self.iterateOnIndex(index)]
		heights = [widget.height() for widget in self.iterateOnIndex(index)]
		return (max(widths), max(heights))
	
	def maxSizes(self):
		return tuple(self.maxSizesByIndex(index) for index in self.columns.indexes())


class TableData(qtw.QWidget):

	showData = qtc.pyqtSignal(dict)

	def __init__(self, model: database_models.TableModel):
		super().__init__()
		self.model = model
		self.selectAll = qtw.QCheckBox()
		self.headers = (self.selectAll,) + tuple(Header(column) for column in self.model.columns)
		self.filters = tuple(self.buildFilter(column) for column in self.model.columns)
		self.main_layout = qtw.QVBoxLayout()
		self.main_layout.setSpacing(0)
		self.setLayout(self.main_layout)
		self.headers_row = qtw.QWidget()
		self.headers_layout = resizable_grid.GridLayout()
		self.headers_layout.setSpacing(2)
		self.headers_layout.setContentsMargins(2, 2, 2, 2)
		self.filterObjects = []
		for i in range(len(self.headers)):
			if i == 0:#selectAll checkbox
				self.headers_layout.addWidget(self.headers[i], 0, i)#, 1, 2)#spans on the headers' and filters' columns
			else:
				self.headers_layout.addWidget(self.headers[i], 0, i)
				self.headers[i].sort.connect(self._onSort)
		for i, filter_widget in enumerate(self.filters):
			self.headers_layout.addWidget(filter_widget, 1, i+1)
			self.connectFilter(filter_widget, i)
		self.headers_row.setLayout(self.headers_layout)
		self.main_layout.addWidget(self.headers_row)
		#now we hide the primary key header and filter:
		self.headers[self.model.columns.primaryKey().index() + 1].hide()#+1because of the selectAllBox
		self.filters[self.model.columns.primaryKey().index()].hide()
		self.dataContainer = qtw.QScrollArea()
		self.dataContainer.setHorizontalScrollBarPolicy(qtc.Qt.ScrollBarAlwaysOff)
		self._rows = None
		self.fillTable()
		self.selectAll.clicked.connect(self._onSelectAll)
		self.main_layout.addWidget(self.dataContainer)
		
	def _onSelectAll(self):
		selected = self.selectAll.isChecked()
		for row in self._rows:
			row.setSelected(selected)
		
	def connectFilter(self, filter_widget, column_index):
		if filter_widget.__class__ == database_filters.TextFilterWidget:
			filter_widget.filterChanged.connect(lambda relation, string: self.applyTextFilter(column_index, relation, string))
		elif filter_widget.__class__ == database_filters.SexFilterWidget:
			filter_widget.filterChanged.connect(lambda relation, sex: self.applySexFilter(column_index, relation, sex))
		elif filter_widget.__class__ == database_filters.DtStringFilterWidget:
			filter_widget.filterChanged.connect(lambda year, month, day, relation, enabled: self.applyDtStringFilter(column_index, year, month, day, relation, enabled))

	def removePreviousFilters(self, column_index):
		if type(column_index) == str:
			column_index = self.columns[column_index].index()
		self.filterObjects = list(filter(lambda filt: filt.column != column_index, self.filterObjects))

	def removePreviousDtStringFilters(self, column_index, relation):
		if type(column_index) == str:
			column_index = self.columns[column_index].index()
		self.filterObjects = list(filter(lambda filt: not (filt.column == column_index and filt.relation == relation), self.filterObjects))
			
	def applyTextFilter(self, column_index, relation, string):
		self.removePreviousFilters(column_index)
		if relation != -1:	
			new_filter = database_filters.Filter(column_index, relation, string)
			self.filterObjects.append(new_filter)
		self.fillTable()
			
	def applySexFilter(self, column_index, relation, sex):
		self.removePreviousFilters(column_index)
		if relation != -1:
			new_filter = database_filters.Filter(column_index, relation, sex)
			self.filterObjects.append(new_filter)
		self.fillTable()

	def applyDtStringFilter(self, column_index, year, month, day, relation, enabled):
		self.removePreviousDtStringFilters(column_index, relation)
		if enabled == 1:
			datetime = qtc.QDateTime(qtc.QDate(year, month, day), qtc.QTime(0, 0, 0))
			new_filter = database_filters.Filter(column_index, relation, datetime)
			self.filterObjects.append(new_filter)
		self.fillTable()
		
	def updateTable(self, values):
		self.model.updateRecord(**values)
		
	def showRecord(self, values_dict):
		self.showData.emit(values_dict)
	
	def	rows(self):
		rows = RowsSequence(tuple(Row(record) for record in self.model.select()), self.model.columns)
		for row in rows:
			row.updated.connect(self.updateTable)
			row.showData.connect(self.showRecord)
		return rows

	def buildFilter(self, column):
		if column.fieldClass() == database_models.Field:
			ret_widget = database_filters.TextFilterWidget()
		elif column.fieldClass() == database_models.SexField:
			ret_widget = database_filters.SexFilterWidget()
		elif column.fieldClass() == database_models.DtStringField:
			ret_widget = database_filters.DtStringFilterWidget()
		else:
			ret_widget = database_filters.FilterWidget()
		return ret_widget

	def _onSort(self, column_name, is_reverse):
		self.model.sortByField(column_name, reverse = is_reverse)
		self.fillTable()

	def fillTable(self):
		self.model.filters = []
		self.model.filters.extend(self.filterObjects)
		if self.dataContainer.widget() != None:
			self.dataContainer.takeWidget()
		
		data_table = qtw.QWidget()
		data_layout = resizable_grid.GridLayout()
		data_layout.setSpacing(2)
		data_layout.setContentsMargins(2, 2, 2, 2)
		
		self._rows = self.rows()
		for row_index, row in enumerate(self._rows):
			for field_index, widget in enumerate(row.widgets):
				data_layout.addWidget(widget, row_index, field_index)
				#now we hide the primary key header:
				if widget.__class__ == delegates.IdLabel:
					widget.hide()
		data_table.setLayout(data_layout)
		
		self.dataContainer.setWidget(data_table)
		
	def adaptSizes(self):
		grid_layout = self.dataContainer.widget().layout()
		grid_layout.fillSpaces()
		self.headers_layout.fillSpaces(self.headers_layout.rowCount(), grid_layout.columnCount())
		for column_index in range(grid_layout.columnCount()):
			max_width = grid_layout.getMaxWidth(column_index)
			grid_layout.setWidth(column_index, max_width)
			self.headers_layout.setWidth(column_index, max_width)
	
	def showEvent(self, event):
		super().showEvent(event)
		
	def addRow(self, data):
		ret = self.model.addRecord(**data)
		self.fillTable()
		return ret
		
	def addVoidRow(self):
		ret = self.model.addVoidRecord()
		self.fillTable()
		return ret
	
	def removeSelectedRows(self):
		selected_rows_id = tuple(row.record.primaryKey() for row in self._rows if row.isSelected())
		ret = True
		for row_id in selected_rows_id:
			ret = self.model.deleteRecord(row_id)
		self.fillTable()
		return ret
		

class TableView(qtw.QMainWindow):

	showData = qtc.pyqtSignal(dict)

	def __init__(self, model: database_models.TableModel):
		super().__init__()
		self.toolbar = self.addToolBar(self.tr("actions"))
		add_action = self.toolbar.addAction(qtg.QIcon(':/icon/images/plus.png'), self.tr("add"))
		remove_action = self.toolbar.addAction(qtg.QIcon(':/icon/images/minus.png'), self.tr("remove"))
		self.tableData = TableData(model)
		self.tableData.showData.connect(self._onShowData)
		self.setCentralWidget(self.tableData)
		add_action.triggered.connect(self.tableData.addVoidRow)
		remove_action.triggered.connect(self.tableData.removeSelectedRows)

	def _onShowData(self, data_dict):
		self.showData.emit(data_dict)


	def showEvent(self, event):
		super().showEvent(event)
		self.tableData.adaptSizes()
		self.setMaximumWidth(self.tableData.dataContainer.widget().width())
